@@ -5,17 +5,22 @@ import logging
 from pathlib import Path
 from typing import Any, Dict, List, Literal, Optional, Union
 from uuid import uuid4
+import requests
 
-from pycrdt import Array, Doc, Map
+from pycrdt import Array, Map
 from pydantic import BaseModel
+from sidecar import Sidecar
 from ypywidgets.comm import CommWidget
 
-from .objects import (
+from jupytergis_core.schema import (
     IGeoJSONSource,
+    IGeoParquetSource,
     IGeoTiffSource,
+    IHeatmapLayer,
     IHillshadeLayer,
     IImageLayer,
     IImageSource,
+    IRasterDemSource,
     IRasterLayer,
     IRasterSource,
     IVectorLayer,
@@ -26,7 +31,6 @@ from .objects import (
     LayerType,
     SourceType,
 )
-from .utils import get_source_layer_names, normalize_path
 
 logger = logging.getLogger(__file__)
 
@@ -41,13 +45,12 @@ class GISDocument(CommWidget):
     """
     Create a new GISDocument object.
 
-    :param path: the path to the file that you would like to open.
-    If not provided, a new empty document will be created.
+    :param path: the path to the file that you would like to open. If not provided, a new ephemeral widget will be created.
     """
 
     def __init__(
         self,
-        path: Optional[str] = None,
+        path: Optional[str | Path] = None,
         latitude: Optional[float] = None,
         longitude: Optional[float] = None,
         zoom: Optional[float] = None,
@@ -56,13 +59,14 @@ class GISDocument(CommWidget):
         pitch: Optional[float] = None,
         projection: Optional[str] = None,
     ):
-        comm_metadata = GISDocument._path_to_comm(path)
-
-        ydoc = Doc()
+        if isinstance(path, Path):
+            path = str(path)
 
         super().__init__(
-            comm_metadata=dict(ymodel_name="@jupytergis:widget", **comm_metadata),
-            ydoc=ydoc,
+            comm_metadata={
+                "ymodel_name": "@jupytergis:widget",
+                **self._make_comm(path=path),
+            }
         )
 
         self.ydoc["layers"] = self._layers = Map()
@@ -71,21 +75,20 @@ class GISDocument(CommWidget):
         self.ydoc["layerTree"] = self._layerTree = Array()
         self.ydoc["metadata"] = self._metadata = Map()
 
-        if path is None:
-            if latitude is not None:
-                self._options["latitude"] = latitude
-            if longitude is not None:
-                self._options["longitude"] = longitude
-            if extent is not None:
-                self._options["extent"] = extent
-            if zoom is not None:
-                self._options["zoom"] = zoom
-            if bearing is not None:
-                self._options["bearing"] = bearing
-            if pitch is not None:
-                self._options["pitch"] = pitch
-            if projection is not None:
-                self._options["projection"] = projection
+        if latitude is not None:
+            self._options["latitude"] = latitude
+        if longitude is not None:
+            self._options["longitude"] = longitude
+        if extent is not None:
+            self._options["extent"] = extent
+        if zoom is not None:
+            self._options["zoom"] = zoom
+        if bearing is not None:
+            self._options["bearing"] = bearing
+        if pitch is not None:
+            self._options["pitch"] = pitch
+        if projection is not None:
+            self._options["projection"] = projection
 
     @property
     def layers(self) -> Dict:
@@ -101,16 +104,39 @@ class GISDocument(CommWidget):
         """
         return self._layerTree.to_py()
 
-    def export_to_qgis(self, path: str) -> bool:
+    def sidecar(
+        self,
+        *,
+        title: str = "JupyterGIS sidecar",
+        anchor: Literal[
+            "split-right",
+            "split-left",
+            "split-top",
+            "split-bottom",
+            "tab-before",
+            "tab-after",
+            "right",
+        ] = "split-right",
+    ):
+        """Open the document in a new sidecar panel.
+
+        :param anchor: Where to position the new sidecar panel.
+        """
+        sidecar = Sidecar(title=title, anchor=anchor)
+        with sidecar:
+            display(self)
+
+    def export_to_qgis(self, path: str | Path) -> bool:
         # Lazy import, jupytergis_qgis of qgis may not be installed
         from jupytergis_qgis.qgis_loader import export_project_to_qgis
 
-        virtual_file = {
-            "layers": self._layers.to_py(),
-            "sources": self._sources.to_py(),
-            "layerTree": reversed_tree(self._layerTree.to_py()),
-            "options": self._options.to_py(),
-        }
+        if isinstance(path, Path):
+            path = str(path)
+
+        virtual_file = self.to_py()
+        virtual_file["layerTree"] = reversed_tree(virtual_file["layerTree"])
+        del virtual_file["metadata"]
+
         return export_project_to_qgis(path, virtual_file)
 
     def add_raster_layer(
@@ -158,11 +184,9 @@ class GISDocument(CommWidget):
         self,
         url: str,
         name: str = "Vector Tile Layer",
-        source_layer: str | None = None,
         attribution: str = "",
         min_zoom: int = 0,
         max_zoom: int = 24,
-        type: Literal["circle", "fill", "line"] = "line",
         color_expr=None,
         opacity: float = 1,
         logical_op: str | None = None,
@@ -175,15 +199,9 @@ class GISDocument(CommWidget):
 
         :param name: The name that will be used for the object in the document.
         :param url: The tiles url.
-        :param source_layer: The source layer to use.
         :param attribution: The attribution.
         :param opacity: The opacity, between 0 and 1.
         """
-        source_layers = get_source_layer_names(url)
-        if source_layer is None and len(source_layers) == 1:
-            source_layer = source_layers[0]
-        if source_layer not in source_layers:
-            raise ValueError(f"source_layer should be one of {source_layers}")
 
         source = {
             "type": SourceType.VectorTileSource,
@@ -208,9 +226,7 @@ class GISDocument(CommWidget):
             "visible": True,
             "parameters": {
                 "source": source_id,
-                "type": type,
                 "opacity": opacity,
-                "sourceLayer": source_layer,
                 "color": color_expr,
                 "opacity": opacity,
             },
@@ -226,44 +242,50 @@ class GISDocument(CommWidget):
 
     def add_geojson_layer(
         self,
-        path: str | None = None,
+        path: str | Path | None = None,
         data: Dict | None = None,
         name: str = "GeoJSON Layer",
-        type: "circle" | "fill" | "line" = "line",
         opacity: float = 1,
         logical_op: str | None = None,
         feature: str | None = None,
         operator: str | None = None,
-        value: Union[str, number, float] | None = None,
+        value: Union[str, int, float] | None = None,
         color_expr=None,
     ):
         """
         Add a GeoJSON Layer to the document.
 
         :param name: The name that will be used for the object in the document.
-        :param path: The path to the JSON file to embed into the jGIS file.
+        :param path: The path to the JSON file or URL to embed into the jGIS file.
         :param data: The raw GeoJSON data to embed into the jGIS file.
-        :param type: The type of the vector layer to create.
-        :param color: The color to apply to features.
         :param opacity: The opacity, between 0 and 1.
         :param color_expr: The style expression used to style the layer, defaults to None
         """
+        if isinstance(path, Path):
+            path = str(path)
+
         if path is None and data is None:
             raise ValueError("Cannot create a GeoJSON layer without data")
 
         if path is not None and data is not None:
             raise ValueError("Cannot set GeoJSON layer data and path at the same time")
 
+        parameters = {}
+
         if path is not None:
-            # We cannot put the path to the file in the model
-            # We don't know where the kernel runs/live
-            # The front-end would have no way of finding the file reliably
-            # TODO Support urls to JSON files, in that case, don't embed the data
-            with open(path, "r") as fobj:
-                parameters = {"data": json.loads(fobj.read())}
+            if path.startswith("http://") or path.startswith("https://"):
+                response = requests.get(path)
+                response.raise_for_status()
+                parameters["path"] = path
+            else:
+                # We cannot put the path to the file in the model
+                # We don't know where the kernel runs/live
+                # The front-end would have no way of finding the file reliably
+                with open(path, "r") as fobj:
+                    parameters["data"] = json.load(fobj)
 
         if data is not None:
-            parameters = {"data": data}
+            parameters["data"] = data
 
         source = {
             "type": SourceType.GeoJSONSource,
@@ -279,7 +301,6 @@ class GISDocument(CommWidget):
             "visible": True,
             "parameters": {
                 "source": source_id,
-                "type": type,
                 "color": color_expr,
                 "opacity": opacity,
             },
@@ -315,7 +336,7 @@ class GISDocument(CommWidget):
         source = {
             "type": SourceType.ImageSource,
             "name": f"{name} Source",
-            "parameters": {"url": url, "coordinates": coordinates},
+            "parameters": {"path": url, "coordinates": coordinates},
         }
 
         source_id = self._add_source(OBJECT_FACTORY.create_source(source, self))
@@ -331,9 +352,9 @@ class GISDocument(CommWidget):
 
     def add_video_layer(
         self,
-        urls: [],
+        urls: List,
         name: str = "Image Layer",
-        coordinates: [] = [],
+        coordinates: Optional[List] = None,
         opacity: float = 1,
     ):
         """
@@ -344,6 +365,8 @@ class GISDocument(CommWidget):
         :param coordinates: Corners of video specified in longitude, latitude pairs.
         :param opacity: The opacity, between 0 and 1.
         """
+        if coordinates is None:
+            coordinates = []
 
         if urls is None or coordinates is None:
             raise ValueError("URLs and Coordinates are required")
@@ -380,14 +403,14 @@ class GISDocument(CommWidget):
         """
         Add a tiff layer
 
-        :param str url: URL of the tif
-        :param int min: Minimum pixel value to be displayed, defaults to letting the map display set the value
-        :param int max: Maximum pixel value to be displayed, defaults to letting the map display set the value
-        :param str name: The name that will be used for the object in the document, defaults to "Tiff Layer"
-        :param bool normalize: Select whether to normalize values between 0..1, if false than min/max have no effect, defaults to True
-        :param bool wrapX: Render tiles beyond the tile grid extent, defaults to False
-        :param float opacity: The opacity, between 0 and 1, defaults to 1.0
-        :param _type_ color_expr: The style expression used to style the layer, defaults to None
+        :param url: URL of the tif
+        :param min: Minimum pixel value to be displayed, defaults to letting the map display set the value
+        :param max: Maximum pixel value to be displayed, defaults to letting the map display set the value
+        :param name: The name that will be used for the object in the document, defaults to "Tiff Layer"
+        :param normalize: Select whether to normalize values between 0..1, if false than min/max have no effect, defaults to True
+        :param wrapX: Render tiles beyond the tile grid extent, defaults to False
+        :param opacity: The opacity, between 0 and 1, defaults to 1.0
+        :param color_expr: The style expression used to style the layer, defaults to None
         """
 
         source = {
@@ -414,6 +437,197 @@ class GISDocument(CommWidget):
 
         return self._add_layer(OBJECT_FACTORY.create_layer(layer, self))
 
+    def add_hillshade_layer(
+        self,
+        url: str,
+        name: str = "Hillshade Layer",
+        urlParameters: Optional[Dict] = None,
+        attribution: str = "",
+    ):
+        """
+        Add a hillshade layer
+
+        :param str url: URL of the hillshade layer
+        :param str name: The name that will be used for the object in the document, defaults to "Hillshade Layer"
+        :param attribution: The attribution.
+        """
+        if urlParameters is None:
+            urlParameters = {}
+
+        source = {
+            "type": SourceType.RasterDemSource,
+            "name": f"{name} Source",
+            "parameters": {
+                "url": url,
+                "attribution": attribution,
+                "urlParameters": urlParameters,
+            },
+        }
+        source_id = self._add_source(OBJECT_FACTORY.create_source(source, self))
+
+        layer = {
+            "type": LayerType.HillshadeLayer,
+            "name": name,
+            "visible": True,
+            "parameters": {"source": source_id},
+        }
+
+        return self._add_layer(OBJECT_FACTORY.create_layer(layer, self))
+
+    def add_heatmap_layer(
+        self,
+        feature: str,
+        path: str | Path | None = None,
+        data: Dict | None = None,
+        name: str = "Heatmap Layer",
+        opacity: float = 1,
+        blur: int = 15,
+        radius: int = 8,
+        gradient: Optional[List[str]] = None,
+    ):
+        """
+        Add a Heatmap Layer to the document.
+
+        :param name: The name that will be used for the object in the document.
+        :param path: The path to the JSON file to embed into the jGIS file.
+        :param data: The raw GeoJSON data to embed into the jGIS file.
+        :param gradient: The color gradient to apply.
+        :param opacity: The opacity, between 0 and 1.
+        :param blur: The blur size in pixels
+        :param radius: The radius size in pixels
+        :param feature: The feature to use to heatmap weights
+        """
+        if isinstance(path, Path):
+            path = str(path)
+
+        if path is None and data is None:
+            raise ValueError("Cannot create a GeoJSON source without data")
+
+        if path is not None and data is not None:
+            raise ValueError("Cannot set GeoJSON source data and path at the same time")
+
+        if path is not None:
+            # We cannot put the path to the file in the model
+            # We don't know where the kernel runs/live
+            # The front-end would have no way of finding the file reliably
+            # TODO Support urls to JSON files, in that case, don't embed the data
+            with open(path, "r") as fobj:
+                parameters = {"data": json.loads(fobj.read())}
+
+        if data is not None:
+            parameters = {"data": data}
+
+        if gradient is None:
+            gradient = ["#00f", "#0ff", "#0f0", "#ff0", "#f00"]
+
+        source = {
+            "type": SourceType.GeoJSONSource,
+            "name": f"{name} Source",
+            "parameters": parameters,
+        }
+
+        source_id = self._add_source(OBJECT_FACTORY.create_source(source, self))
+
+        layer = {
+            "type": LayerType.HeatmapLayer,
+            "name": name,
+            "visible": True,
+            "parameters": {
+                "source": source_id,
+                "type": type,
+                "color": gradient,
+                "opacity": opacity,
+                "blur": blur,
+                "radius": radius,
+                "feature": feature,
+            },
+        }
+
+        return self._add_layer(OBJECT_FACTORY.create_layer(layer, self))
+
+    def add_geoparquet_layer(
+        self,
+        path: str,
+        name: str = "GeoParquetLayer",
+        type: "circle" | "fill" | "line" = "line",
+        opacity: float = 1,
+        logical_op: str | None = None,
+        feature: str | None = None,
+        operator: str | None = None,
+        value: Union[str, int, float] | None = None,
+        color_expr=None,
+    ):
+        """
+        Add a GeoParquet Layer to the document
+        :param path: The path to the GeoParquet file to embed into the jGIS file.
+        :param name: The name that will be used for the object in the document.
+        :param type: The type of the vector layer to create.
+        :param opacity: The opacity, between 0 and 1.
+        :param logical_op: The logical combination to apply to filters. Must be "any" or "all"
+        :param feature: The feature to be filtered on
+        :param operator: The operator used to compare the feature and value
+        :param value: The value to be filtered on
+        :param color_expr: The style expression used to style the layer
+        """
+
+        source = {
+            "type": SourceType.GeoParquetSource,
+            "name": f"{name} Source",
+            "parameters": {"path": path},
+        }
+
+        source_id = self._add_source(OBJECT_FACTORY.create_source(source, self))
+
+        layer = {
+            "type": LayerType.VectorLayer,
+            "name": name,
+            "visible": True,
+            "parameters": {
+                "source": source_id,
+                "type": type,
+                "opacity": opacity,
+                "color": color_expr,
+            },
+            "filters": {
+                "appliedFilters": [
+                    {"feature": feature, "operator": operator, "value": value}
+                ],
+                "logicalOp": logical_op,
+            },
+        }
+
+        return self._add_layer(OBJECT_FACTORY.create_layer(layer, self))
+
+    def remove_layer(self, layer_id: str):
+        """
+        Remove a layer from the GIS document.
+
+        :param layer_id: The ID of the layer to remove.
+        :raises KeyError: If the layer does not exist.
+        """
+
+        layer = self._layers.get(layer_id)
+
+        if layer is None:
+            raise KeyError(f"No layer found with ID: {layer_id}")
+
+        del self._layers[layer_id]
+        self._remove_source_if_orphaned(layer["parameters"]["source"])
+
+    def _remove_source_if_orphaned(self, source_id: str):
+        source = self._sources.get(source_id)
+
+        if source is None:
+            raise KeyError(f"No source found with ID: {source_id}")
+
+        source_is_orphan = not any(
+            layer["parameters"]["source"] == source_id
+            for layer in self._layers.values()
+        )
+
+        if source_is_orphan:
+            del self._sources[source_id]
+
     def create_color_expr(
         self,
         color_stops: Dict,
@@ -473,16 +687,16 @@ class GISDocument(CommWidget):
         logical_op: str,
         feature: str,
         operator: str,
-        value: Union[str, number, float],
+        value: Union[str, int, float],
     ):
         """
         Add a filter to a layer
 
-        :param str layer_id: The ID of the layer to filter
-        :param str logical_op: The logical combination to apply to filters. Must be "any" or "all"
-        :param str feature: The feature to be filtered on
-        :param str operator: The operator used to compare the feature and value
-        :param Union[str, number, float] value: The value to be filtered on
+        :param layer_id: The ID of the layer to filter
+        :param logical_op: The logical combination to apply to filters. Must be "any" or "all"
+        :param feature: The feature to be filtered on
+        :param operator: The operator used to compare the feature and value
+        :param value: The value to be filtered on
         """
         layer = self._layers.get(layer_id)
 
@@ -519,16 +733,16 @@ class GISDocument(CommWidget):
         logical_op: str,
         feature: str,
         operator: str,
-        value: Union[str, number, float],
+        value: Union[str, int, float],
     ):
         """
         Update a filter applied to a layer
 
-        :param str layer_id: The ID of the layer to filter
-        :param str logical_op: The logical combination to apply to filters. Must be "any" or "all"
-        :param str feature: The feature to update the value for
-        :param str operator: The operator used to compare the feature and value
-        :param Union[str, number, float] value: The new value to be filtered on
+        :param layer_id: The ID of the layer to filter
+        :param logical_op: The logical combination to apply to filters. Must be "any" or "all"
+        :param feature: The feature to update the value for
+        :param operator: The operator used to compare the feature and value
+        :param value: The new value to be filtered on
         """
         layer = self._layers.get(layer_id)
 
@@ -576,13 +790,13 @@ class GISDocument(CommWidget):
         layer["filters"]["appliedFilters"] = []
         self._layers[layer_id] = layer
 
-    def _add_source(self, new_object: "JGISObject"):
-        _id = str(uuid4())
+    def _add_source(self, new_object: "JGISObject", id: str | None = None) -> str:
+        _id = str(uuid4()) if id is None else id
         obj_dict = json.loads(new_object.json())
         self._sources[_id] = obj_dict
         return _id
 
-    def _add_layer(self, new_object: "JGISObject"):
+    def _add_layer(self, new_object: "JGISObject") -> str:
         _id = str(uuid4())
         obj_dict = json.loads(new_object.json())
         self._layers[_id] = obj_dict
@@ -590,18 +804,17 @@ class GISDocument(CommWidget):
         return _id
 
     @classmethod
-    def _path_to_comm(cls, filePath: Optional[str]) -> Dict:
-        path = None
+    def _make_comm(cls, *, path: Optional[str]) -> Dict:
         format = None
         contentType = None
 
-        if filePath is not None:
-            path = normalize_path(filePath)
+        if path is not None:
             file_name = Path(path).name
             try:
                 ext = file_name.split(".")[1].lower()
-            except Exception:
-                raise ValueError("Can not detect file extension!")
+            except Exception as e:
+                raise ValueError("Can not detect file extension!") from e
+
             if ext == "jgis":
                 format = "text"
                 contentType = "jgis"
@@ -613,10 +826,24 @@ class GISDocument(CommWidget):
                 contentType = "QGS"
             else:
                 raise ValueError("File extension is not supported!")
+
         return dict(
-            path=path, format=format, contentType=contentType, createydoc=path is None
+            path=path,
+            format=format,
+            contentType=contentType,
+            create_ydoc=path is None,
         )
 
+    def to_py(self) -> dict:
+        """Get the document structure as a Python dictionary."""
+        return {
+            "layers": self._layers.to_py(),
+            "sources": self._sources.to_py(),
+            "layerTree": self._layerTree.to_py(),
+            "options": self._options.to_py(),
+            "metadata": self._metadata.to_py(),
+        }
+
 
 class JGISLayer(BaseModel):
     class Config:
@@ -633,6 +860,7 @@ class JGISLayer(BaseModel):
         IHillshadeLayer,
         IImageLayer,
         IWebGlLayer,
+        IHeatmapLayer,
     ]
     _parent = Optional[GISDocument]
 
@@ -655,6 +883,8 @@ class JGISSource(BaseModel):
         IImageSource,
         IVideoSource,
         IGeoTiffSource,
+        IRasterDemSource,
+        IGeoParquetSource,
     ]
     _parent = Optional[GISDocument]
 
@@ -733,6 +963,7 @@ OBJECT_FACTORY.register_factory(LayerType.VectorTileLayer, IVectorTileLayer)
 OBJECT_FACTORY.register_factory(LayerType.HillshadeLayer, IHillshadeLayer)
 OBJECT_FACTORY.register_factory(LayerType.WebGlLayer, IWebGlLayer)
 OBJECT_FACTORY.register_factory(LayerType.ImageLayer, IImageLayer)
+OBJECT_FACTORY.register_factory(LayerType.HeatmapLayer, IHeatmapLayer)
 
 OBJECT_FACTORY.register_factory(SourceType.VectorTileSource, IVectorTileSource)
 OBJECT_FACTORY.register_factory(SourceType.RasterSource, IRasterSource)
@@ -740,3 +971,5 @@ OBJECT_FACTORY.register_factory(SourceType.GeoJSONSource, IGeoJSONSource)
 OBJECT_FACTORY.register_factory(SourceType.ImageSource, IImageSource)
 OBJECT_FACTORY.register_factory(SourceType.VideoSource, IVideoSource)
 OBJECT_FACTORY.register_factory(SourceType.GeoTiffSource, IGeoTiffSource)
+OBJECT_FACTORY.register_factory(SourceType.RasterDemSource, IRasterDemSource)
+OBJECT_FACTORY.register_factory(SourceType.GeoParquetSource, IGeoParquetSource)
