@@ -7,17 +7,21 @@ logger = logging.getLogger(__file__)
 
 
 class YDocConnector(Widget):
-    def __init__(self, path: Optional[str], **kwargs) -> None:
+    def __init__(self, path: Optional[str | Path], **kwargs) -> None:
         self.path = None
         self._format = None
         self._contentType = None
 
+        if isinstance(path, Path):
+            path = str(path)
+
         if path is not None:
             self.path = path
             try:
                 ext = path.split(".")[1].lower()
-            except Exception:
-                raise Exception("Can not detect file extension!")
+            except Exception as e:
+                raise Exception("Can not detect file extension!") from e
+
             if ext == "jgis":
                 self._format = "text"
                 self._contentType = "jgis"
